#!/usr/bin/env python
"""
Functions for running dyPolyChord using PyPolyChord (PolyChord's built-in
python wrapper) using with python likelihoods and priors.
"""
# Exception handling needed to allow readthedocs to work without installing
# PyPolyChord.
try:
    import PyPolyChord
    import PyPolyChord.settings
except ImportError:
    pass


class RunPyPolyChord(object):

    """Callable class for running PolyChord in Python with specified
    the settings."""

    def __init__(self, likelihood, prior, ndim, nderived=0):
        """
        Specify likelihood, prior and number of dimensions in calculation.

        Parameters
        ----------
        likelihood: func
        prior: func
        ndim: int
        nderived: int, optional
        """
        self.likelihood = likelihood
        self.prior = prior
        self.ndim = ndim
        self.nderived = nderived

    def __call__(self, settings_dict, comm=None):
        """
        Runs PyPolyChord with specified inputs and writes output files. See the
        PyPolyChord documentation for more details.

        Parameters
        ----------
        settings_dict: dict
            Input PolyChord settings.
        comm: None or mpi4py MPI.COMM object, optional
            For MPI parallelisation.
        """
        if comm is None:
            settings = PyPolyChord.settings.PolyChordSettings(
                self.ndim, self.nderived, **settings_dict)
        else:
            rank = comm.Get_rank()
            if rank == 0:
                settings = PyPolyChord.settings.PolyChordSettings(
                    self.ndim, self.nderived, **settings_dict)
            else:
                settings = None
            settings = comm.bcast(settings, root=0)
        PyPolyChord.run_polychord(self.likelihood, self.ndim, self.nderived,
                                  settings, prior=self.prior)



#!/usr/bin/env python
"""
Functions for running dyPolyChord using compiled PolyChord C++ or Fortran
likelihoods.
"""
import os


def get_prior_block_str(prior_name, prior_params, nparam, **kwargs):
    """
    Returns a PolyChord format prior block for inclusion in PolyChord .ini
    files.

    See the PolyChord documentation for more details.

    Parameters
    ----------
    prior_name: str
        Name of prior. See the PolyChord documentation for a list of currently
        available priors and details of how to add your own.
    prior_params: str, float or list of strs and floats
        Parameters for the prior function.
    nparam: int
        Number of parameters.
    start_param: int, optional
        Where to start param numbering. For when multiple prior blocks are
        being used.
    block: int, optional
        Number of block (only needed when using multiple prior blocks).
    speed: int, optional
        Use to specify fast and slow parameters if required. See the PolyChord
        documentation for more details.

    Returns
    -------
    block_str: str
        PolyChord format prior block.
    """
    start_param = kwargs.pop('start_param', 1)
    speed = kwargs.pop('speed', 1)
    block = kwargs.pop('block', 1)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    block_str = ''
    for i in range(start_param, nparam + start_param):
        block_str += ('P : p{0} | \\theta_{{{0}}} | {1} | {2} | {3} |'
                      .format(i, speed, prior_name, block))
        block_str += format_setting(prior_params) + '\n'
    return block_str


class RunCompiledPolyChord(object):

    """Object for running a compiled PolyChord executable with specified
    inputs."""

    def __init__(self, executable_path, prior_str, derived_str=None,
                 mpi_str=None):
        """
        Specify path to executable, priors and derived parameters.

        Parameters
        ----------
        executable_path: str
            Path to compiled likelihood. If this is in the directory from which
            dyPolyChord is being run, you may need to prepend "./" to the
            executable name for it to work.
        prior_str: str
            String specifying prior in the format required for PolyChord .ini
            files (see get_prior_block_str for more details).
        derived_str: str or None, optional
            String specifying prior in the format required for PolyChord .ini
            files (see prior_str for more details).
        mpi_str: str or None, optional
            Optional mpi command to preprend to run command.
            For example to run with 8 processors, use mpi_str = 'mprun -np 8'.
            Note that PolyChord must be installed with MPI enabled to allow
            running with MPI.
        """
        self.executable_path = executable_path
        self.prior_str = prior_str
        self.derived_str = derived_str
        self.mpi_str = mpi_str

    def __call__(self, settings_dict, comm=None):
        """
        Run PolyChord with the input settings by writing a .ini file then using
        the compiled likelihood specified in executable_path.

        See the PolyChord documentation for more details.

        Parameters
        ----------
        settings_dict: dict
            Input PolyChord settings.
        comm: None, optional
            Not used. Included only so __call__ has the same arguments as the
            equivalent python function (which uses the comm argument for
            runnign with MPI).
        """
        assert os.path.isfile(self.executable_path)
        assert comm is None, 'comm not used for compiled likelihoods.'
        ini_path = os.path.join(settings_dict['base_dir'],
                                settings_dict['file_root'] + '.ini')
        self.write_ini(settings_dict, ini_path)
        command_str = self.executable_path + ' ' + ini_path
        if self.mpi_str is not None:
            command_str = self.mpi_str + ' ' + command_str
        os.system(command_str)

    def write_ini(self, settings, file_path):
        """
        Writes a PolyChord format .ini file based on the input settings.

        Parameters
        ----------
        settings: dict
            Input PolyChord settings.
        file_path: str
            Path to write ini file to.
        """
        with open(file_path, 'w') as ini_file:
            # Write the settings
            for key, value in settings.items():
                if key == 'nlives':
                    if value:
                        loglikes = sorted(settings['nlives'])
                        ini_file.write(
                            'loglikes = ' + format_setting(loglikes) + '\n')
                        nlives = [settings['nlives'][ll] for ll in loglikes]
                        ini_file.write(
                            'nlives = ' + format_setting(nlives) + '\n')
                else:
                    ini_file.write(key + ' = ' + format_setting(value) + '\n')
            # write the prior
            ini_file.write(self.prior_str)
            if self.derived_str is not None:
                ini_file.write(self.derived_str)


def format_setting(setting):
    """
    Return setting as string in the format needed for PolyChord's .ini files.
    These use 'T' for True and 'F' for False, and require lists of numbers
    written separated by spaces and without commas or brackets.

    Parameters
    ----------
    setting: (can be any type for which str(settings) works)

    Returns
    -------
    str
    """
    if isinstance(setting, bool):
        return str(setting)[0]
    elif isinstance(setting, (list, tuple)):
        string = str(setting)
        for char in [',', '[', ']', '(', ')']:
            string = string.replace(char, '')
        return string
    else:
        return str(setting)



#!/usr/bin/env python
"""
Functions for loading and processing dyPolyChord dynamic nested sampling runs.
"""
import os
import numpy as np
import nestcheck.ns_run_utils
import nestcheck.data_processing
import nestcheck.io_utils as iou


def settings_root(likelihood_name, prior_name, ndims, **kwargs):
    """
    Returns a standard string containing information about settings.

    Parameters
    ----------
    likelihood_name: str
    prior_name: str
    ndims: int
    prior_scale: float or int
    dynamic_goal: float, int or None
    nlive_const: int
    nrepeats: int
    nint: int, optional
    init_step: int, optional

    Returns
    -------
    root: str
    """
    prior_scale = kwargs.pop('prior_scale')
    dynamic_goal = kwargs.pop('dynamic_goal')
    nlive_const = kwargs.pop('nlive_const')
    nrepeats = kwargs.pop('nrepeats')
    ninit = kwargs.pop('ninit', None)
    init_step = kwargs.pop('init_step', None)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    root = likelihood_name + '_' + prior_name + '_' + str(prior_scale)
    root += '_dg' + str(dynamic_goal)
    if dynamic_goal is not None:
        assert ninit is not None
        root += '_' + str(ninit) + 'init'
        if dynamic_goal != 0:
            assert init_step is not None
            root += '_' + str(init_step) + 'is'
    root += '_' + str(ndims) + 'd'
    root += '_' + str(nlive_const) + 'nlive'
    root += '_' + str(nrepeats) + 'nrepeats'
    return root


def process_dypolychord_run(file_root, base_dir, **kwargs):
    """
    Load the output files of a dynamic run and process them to the nestcheck
    format.

    Parameters
    ----------
    file_root: str
    base_dir: str
    dynamic_goal: float
    logl_warn_only: bool, optional
        Whether to throw error or warning if there are duplicate point
        likelihoods. This is passed to nestcheck's check_ns_run function; see
        its documentation for more details.

    Returns
    -------
    run: dict
        Nested sampling run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information).
    """
    dynamic_goal = kwargs.pop('dynamic_goal')
    logl_warn_only = kwargs.pop('logl_warn_only', False)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    init = nestcheck.data_processing.process_polychord_run(
        file_root + '_init', base_dir, logl_warn_only=logl_warn_only)
    assert np.all(init['thread_min_max'][:, 0] == -np.inf), (
        str(init['thread_min_max']))
    dyn = nestcheck.data_processing.process_polychord_run(
        file_root + '_dyn', base_dir, logl_warn_only=logl_warn_only)
    dyn_info = iou.pickle_load(os.path.join(
        base_dir, file_root + '_dyn_info'))
    if dynamic_goal == 0:
        # If dynamic_goal == 0 nlive only decreases, so check all threads
        # start by sampling
        assert np.all(dyn['thread_min_max'][:, 0] == -np.inf), (
            str(dyn['thread_min_max']))
    if 'resume_ndead' not in dyn_info:
        # If dynamic_goal == 0, dyn was not resumed part way through init:
        # hence there are no samples repeated in both runs' files and we can
        # simply combine dyn and init using standard nestcheck functions.
        run = nestcheck.ns_run_utils.combine_ns_runs([init, dyn])
        nlike = init['output']['nlike'] + dyn['output']['nlike']
    else:
        # If dynamic_goal == 1, dyn was resumed part way through init and we
        # need to remove duplicate points from the combined run
        run = combine_resumed_dyn_run(init, dyn, dyn_info['resume_ndead'])
        nlike = (init['output']['nlike'] + dyn['output']['nlike']
                 - dyn_info['resume_nlike'])
    # Add info to run
    run['output'] = {'nlike': nlike,
                     'file_root': file_root,
                     'base_dir': base_dir}
    # check the nested sampling run has the expected properties
    nestcheck.data_processing.check_ns_run(run, logl_warn_only=logl_warn_only)
    return run


def combine_resumed_dyn_run(init, dyn, resume_ndead):
    """
    Merge initial run and dynamic run which was resumed from it, including
    removing duplicate points present in both runs.

    Parameters
    ----------
    init: dict
        Initial exploratory run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information).
    dyn: dict
        Dynamic run in nestcheck format.
    resume_ndead: int
        The number of dead points present when dyn was resumed from init.

    Returns
    -------
    run: dict
        Combined run in nestcheck format.
    """
    assert np.array_equal(
        init['logl'][:resume_ndead], dyn['logl'][:resume_ndead]), (
            'The first {0} points should be the same'.format(resume_ndead))
    init['theta'] = init['theta'][resume_ndead:, :]
    for key in ['nlive_array', 'logl', 'thread_labels']:
        init[key] = init[key][resume_ndead:]
    # We also need to remove the points that were live when the resume file was
    # written, as these show up as dead points in dyn
    live_inds = []
    empty_thread_inds = []
    for i, th_lab in enumerate(np.unique(init['thread_labels'])):
        th_inds = np.where(init['thread_labels'] == th_lab)[0]
        live_inds.append(th_inds[0])
        live_logl = init['logl'][th_inds[0]]
        if th_inds.shape[0] == 1:
            empty_thread_inds.append(i)
        assert np.where(dyn['logl'] == live_logl)[0].shape == (1,), (
            'point should be present in dyn too! logl=' + str(live_logl))
        init['thread_min_max'][i, 0] = live_logl
    # lets remove the live points at init
    init['theta'] = np.delete(init['theta'], live_inds, axis=0)
    for key in ['nlive_array', 'logl', 'thread_labels']:
        init[key] = np.delete(init[key], live_inds)
    # Deal with the case that one of the threads is now empty
    if empty_thread_inds:
        # remove any empty threads from logl_min_max
        init['thread_min_max'] = np.delete(
            init['thread_min_max'], empty_thread_inds, axis=0)
        # Now we need to reorder the thread labels to avoid gaps
        thread_labels_new = np.full(init['thread_labels'].shape, np.nan)
        for i, th_lab in enumerate(np.unique(init['thread_labels'])):
            inds = np.where(init['thread_labels'] == th_lab)[0]
            thread_labels_new[inds] = i
            # Check the newly relabelled thread label matches thread_min_max
            assert init['thread_min_max'][i, 0] <= init['logl'][inds[0]]
            assert init['thread_min_max'][i, 1] == init['logl'][inds[-1]]
        assert np.all(~np.isnan(thread_labels_new))
        init['thread_labels'] = thread_labels_new.astype(int)
    # Add the init threads to dyn with new labels that continue on from the dyn
    # labels
    init['thread_labels'] += dyn['thread_min_max'].shape[0]
    run = nestcheck.ns_run_utils.combine_threads(
        nestcheck.ns_run_utils.get_run_threads(dyn) +
        nestcheck.ns_run_utils.get_run_threads(init),
        assert_birth_point=True)
    return run



#!/usr/bin/env python
"""
Functions for selecting the number of live points in order to maximise
calculation accuracy.
"""
import warnings
import itertools
import numpy as np
import nestcheck.ns_run_utils
import nestcheck.data_processing


def allocate(init_run, samp_tot, dynamic_goal, smoothing_filter=None):
    """
    Calculates an allocation of life points for dynamic run, checks the output
    allocation and the smoothing applied, and returns the information needed
    for the dynamic run in a dictionary.

    Parameters
    ----------
    init_run: dict
        Initial exploratory run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information).
    samp_tot: int
        Total number of samples (including both the initial exploratory run and
        the second dynamic run).
    dynamic_goal: float
    smoothing_filter: function or None, optional
        Smoothing for nlive array. Set to None for no smoothing.

    Returns
    -------
    dyn_info: dict
    """
    assert np.all(np.diff(init_run['logl']) > 0)
    # Calculate nlive allocation with and without smoothing
    nlives = dyn_nlive_array(init_run, samp_tot, dynamic_goal,
                             smoothing_filter=smoothing_filter)
    nlives_unsmoothed = dyn_nlive_array(init_run, samp_tot, dynamic_goal,
                                        smoothing_filter=None)
    # Perform some checks
    if dynamic_goal == 0:
        if not np.all(np.diff(nlives) <= 0):
            assert np.all(np.diff(nlives_unsmoothed) <= 0)
            warnings.warn((
                'Smoothing has added turning points to nlive allocation when '
                'dynamic_goal=0. I am using the unsmoothed nlives '
                'instead.'), UserWarning)
            nlives = nlives_unsmoothed
        assert nlives[0] == nlives.max(), str(nlives)
        assert np.all(np.diff(nlives) <= 0), (
            'When targeting evidence, nlive should monotincally decrease!'
            + ' nlives = ' + str(nlives))
    if dynamic_goal == 1 and nlives[0] != 0:
        warnings.warn(
            'dynamic_goal=1 but the first allocated nlive point is {0}. '
            'For most likelihoods we expect this to equal zero (although '
            'it may be nonzero if there is significant posterior mass '
            'at the edge of the prior).'.format(nlives[0]), UserWarning)
    assert nlives.min() == 0
    # Find number of blocks where nlive is nonzero
    nonzero_blocks = np.asarray(
        [k for k, g in itertools.groupby(nlives != 0)])
    assert nonzero_blocks.sum() == 1, (
        'nlive becomes zero then becomes nonzero! nonzero_blocks='
        + str(nonzero_blocks))
    # Get the indexes of nlives points which are different to the previous
    # points (i.e. remove consecutive duplicates, keeping first occurance)
    inds_to_use = np.concatenate(
        (np.asarray([0]), np.where(np.diff(nlives) != 0)[0] + 1))
    assert (count_turning_points(nlives) ==
            count_turning_points(nlives[inds_to_use]))
    # Check logl = approx -inf is mapped to the starting number of live points
    nlives_dict = {-1.e100: int(nlives[0])}
    for ind in inds_to_use:
        nlives_dict[init_run['logl'][ind]] = int(nlives[ind])
    # Store the nlive allocations for dyn_info
    dyn_info = {'init_nlive_allocation': nlives,
                'init_nlive_allocation_unsmoothed': nlives_unsmoothed,
                'nlives_dict': nlives_dict,
                'peak_start_ind': np.where(nlives > 0)[0][0]}
    return dyn_info


def dyn_nlive_array(init_run, samp_tot, dynamic_goal, smoothing_filter=None):
    """
    Calculate the dynamic nlive allocation from the theoretical, point
    importance-based allocation. This allows for the samples taken in the
    initial run, including areas where more samples than were needed have been
    taken (meaning there are fewer samples available for the regions with peak
    importance). Also perform optional smoothing (carried out *before*
    rounding).

    Normalisation uses the the formulae for the expected number of samples:

    .. math:: N_\\mathrm{samp} = \\int n(\\log X) \\mathrm{d}\\log X

    where :math:`n` is the local number of live points.

    See Appendix E of "Dynamic nested sampling: an improved algorithm for
    parameter estimation and evidence calculation" (Higson et al., 2018) for
    more information.

    Parameters
    ----------
    init_run: dict
        Initial exploratory run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information).
    samp_tot: int
        Total number of samples (including both the initial exploratory run and
        the second dynamic run).
    dynamic_goal: float
    smoothing_filter: function or None, optional
        Smoothing for nlive array. Set to None for no smoothing.

    Returns
    -------
    nlive_array: 1d numpy array
        Number of live points corresponding to each likelihood in
        init_run['logl'].
    """
    assert samp_tot > init_run['logl'].shape[0]
    # Calculate the importance of each point
    importance = sample_importance(init_run, dynamic_goal)
    # Calculate theoretical nlive allocation, which is proportional to
    # importance and normalised to produce an expected samp_tot samples
    logx_init = nestcheck.ns_run_utils.get_logx(init_run['nlive_array'])
    norm = samp_tot / np.abs(np.trapz(importance, x=logx_init))
    importance_nlive = importance * norm
    # Account for the points already sampled
    importance_nlive -= init_run['nlive_array'][0]
    if smoothing_filter is None:
        nlive_array = importance_nlive
    else:
        nlive_array = smoothing_filter(importance_nlive)
    nlive_array = np.clip(nlive_array, 0, None)
    # Renormalise to account for nlives below zero (i.e. regions where we have
    # already taken too many samples) as we cannot take negative samples.
    samp_remain = samp_tot - init_run['logl'].shape[0]
    nlive_array *= samp_remain / np.abs(np.trapz(nlive_array, x=logx_init))
    return np.rint(nlive_array)


def sample_importance(run, dynamic_goal):
    """
    Calculate the importance of each sample in the run.

    See "Dynamic nested sampling: an improved algorithm for parameter
    estimation and evidence calculation" (Higson et al., 2018) for more
    information.

    Parameters
    ----------
    run: dict
        Nested sampling run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information)
    dynamic_goal: float or int
        Specifies how much computational effort to put into improving evidence
        and parameter estimation accuracy.

    Returns
    -------
    importance: 1d numpy array
        Importance of each sample. Normalised to np.sum(importance) = 1
    """
    assert 0 <= dynamic_goal <= 1, (
        'dynamic_goal={0} not in [0,1]'.format(dynamic_goal))
    logw = run['logl'] + nestcheck.ns_run_utils.get_logx(run['nlive_array'])
    w_rel = np.exp(logw - logw.max())
    param_imp = w_rel / np.sum(w_rel)
    if dynamic_goal == 1:
        return param_imp
    z_imp = np.cumsum(w_rel)
    z_imp = z_imp.max() - z_imp
    z_imp /= np.sum(z_imp)
    assert z_imp[0] == z_imp.max()
    if dynamic_goal == 0:
        return z_imp
    return (dynamic_goal * param_imp) + ((1 - dynamic_goal) * z_imp)


def count_turning_points(array):
    """Returns number of turning points in a 1d array."""
    # Remove adjacent duplicates only
    uniq = np.asarray([k for k, g in itertools.groupby(array)])
    # Count turning points
    return (np.diff(np.sign(np.diff(uniq))) != 0).sum()



#!/usr/bin/env python
"""
Contains main function for running dynamic nested sampling.
"""
import copy
import os
import shutil
import warnings
import numpy as np
import scipy.signal
import nestcheck.data_processing
import nestcheck.io_utils
import nestcheck.write_polychord_output
import dyPolyChord.nlive_allocation
import dyPolyChord.output_processing


__all__ = ['run_dypolychord', 'check_settings']


@nestcheck.io_utils.timing_decorator
def run_dypolychord(run_polychord, dynamic_goal, settings_dict_in, **kwargs):
    """
    Performs dynamic nested sampling using the algorithm described in
    Appendix E of "Dynamic nested sampling: an improved algorithm for
    parameter estimation and evidence calculation" (Higson et al., 2018).
    This proceeds in 3 steps:

    1) Generate an initial run with a constant number of live points
    :math:`n_\\mathrm{init}`. This process is run in chuncks using PolyChord's
    max_ndead setting to allow periodic saving of .resume files so the initial
    run can be resumed at different points.
    2) Calculate an allocation of the number of live points at each likelihood
    for use in step 3. Also cleans up resume files and saves relevant
    information.
    3) Generate dynamic nested sampling run using the calculated live point
    allocation.
    4) Combine the initial and dynamic runs and write output files in the
    PolyChord format, and remove the intermediate output files produced.

    Parameters
    ----------
    run_polychord: callable
        Callable which runs PolyChord with the desired likelihood and prior,
        and takes a settings dictionary as its argument.
    dynamic_goal: float or int
        Number in (0, 1) which determines how to allocate computational effort
        between parameter estimation and evidence calculation. See the dynamic
        nested sampling paper for more details.
    settings_dict: dict
        PolyChord settings to use (see check_settings for information on
        allowed and default settings).
    nlive_const: int, optional
        Used to calculate total number of samples if max_ndead not specified in
        settings. The total number of samples used is the estimated number that
        would be taken by a nested sampling run with a constant number of live
        points nlive_const.
    ninit: int, optional
        Number of live points to use for the initial exporatory run (Step 1).
    ninit_step: int, optional
        Number of samples taken between saving .resume files in Step 1.
    seed_increment: int, optional
        If seeding is used (PolyChord seed setting >= 0), this increment is
        added to PolyChord's random seed each time it is run to avoid
        repeated points.
        When running in parallel using MPI, PolyChord hashes the seed with the
        MPI rank using IEOR. Hence you need seed_increment to be > number of
        processors to ensure no two processes use the same seed.
        When running repeated results you need to increment the seed used for
        each run by some number >> seed_increment.
    smoothing_filter: func, optional
        Smoothing to apply to the nlive allocation (if any).
    logl_warn_only: bool, optional
        Whether to raise error or warning if multiple samples have the same
        loglikelihood. This is passed to nestcheck's check_ns_run function;
        see its documentation for more details.
    stats_means_errs: bool, optional
        Whether to include estimates of logZ and parameter mean values and
        their uncertainties in the .stats file. This is passed to nestcheck's
        write_run_output; see its documentation for more details.
    clean: bool, optional
        Clean the additional output files made by dyPolyChord, leaving only
        output files for the combined run in PolyChord format.
        When debugging this can be set to False to allow inspection of
        intermediate output.
    """
    try:
        nlive_const = kwargs.pop('nlive_const', settings_dict_in['nlive'])
    except KeyError:
        nlive_const = kwargs.pop('nlive_const', 100)
    ninit = kwargs.pop('ninit', 10)
    init_step = kwargs.pop('init_step', ninit)
    seed_increment = kwargs.pop('seed_increment', 100)
    default_smoothing = (lambda x: scipy.signal.savgol_filter(
        x, 1 + (2 * ninit), 3, mode='nearest'))
    smoothing_filter = kwargs.pop('smoothing_filter', default_smoothing)
    comm = kwargs.pop('comm', None)
    logl_warn_only = kwargs.pop('logl_warn_only', False)
    stats_means_errs = kwargs.pop('stats_means_errs', True)
    clean = kwargs.pop('clean', True)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    # Step 1: do initial run
    # ----------------------
    # set up rank if running with MPI
    if comm is not None:
        rank = comm.Get_rank()
    else:
        rank = 0
    settings_dict = None  # define for rank != 0
    if rank == 0:
        settings_dict_in, output_settings = check_settings(
            settings_dict_in)
        # Make a copy of settings dic so we dont edit settings
        settings_dict = copy.deepcopy(settings_dict_in)
        settings_dict['file_root'] = settings_dict['file_root'] + '_init'
        settings_dict['nlive'] = ninit
    if dynamic_goal == 0:
        # We definitely won't need to resume midway through in this case, so
        # just run PolyChod normally
        run_polychord(settings_dict, comm=comm)
        final_seed = settings_dict['seed']
        if settings_dict['seed'] >= 0:
            final_seed += seed_increment
    else:
        step_ndead, resume_outputs, final_seed = run_and_save_resumes(
            run_polychord, settings_dict, init_step, seed_increment, comm=comm)
    # Step 2: calculate an allocation of live points
    # ----------------------------------------------
    if rank == 0:
        init_run = nestcheck.data_processing.process_polychord_run(
            settings_dict_in['file_root'] + '_init',
            settings_dict_in['base_dir'])
        # Calculate max number of samples
        if settings_dict_in['max_ndead'] > 0:
            samp_tot = settings_dict_in['max_ndead']
            assert settings_dict_in['max_ndead'] > init_run['logl'].shape[0], (
                'all points used in init run - None left for dynamic run')
        else:
            samp_tot = init_run['logl'].shape[0] * (nlive_const / ninit)
            assert nlive_const > ninit
        dyn_info = dyPolyChord.nlive_allocation.allocate(
            init_run, samp_tot, dynamic_goal,
            smoothing_filter=smoothing_filter)
        root_name = os.path.join(settings_dict_in['base_dir'],
                                 settings_dict_in['file_root'])
        if dyn_info['peak_start_ind'] != 0:
            # subtract 1 as ndead=1 corresponds to point 0
            resume_steps = np.asarray(step_ndead) - 1
            # Work out which resume file to load
            resume_ndead = step_ndead[np.where(
                resume_steps < dyn_info['peak_start_ind'])[0][-1]]
            # copy resume step to dynamic file root
            shutil.copyfile(
                root_name + '_init_' + str(resume_ndead) + '.resume',
                root_name + '_dyn.resume')
            # Save resume info
            dyn_info['resume_ndead'] = resume_ndead
            dyn_info['resume_nlike'] = resume_outputs[resume_ndead]['nlike']
        nestcheck.io_utils.pickle_save(
            dyn_info, root_name + '_dyn_info', overwrite_existing=True)
        if dynamic_goal != 0:
            # Remove all the temporary resume files. Use set to avoid
            # duplicates as these cause OSErrors.
            for snd in set(step_ndead):
                os.remove(root_name + '_init_' + str(snd) + '.resume')
        # Step 3: do dynamic run
        # ----------------------
        # remove edits from init
        settings_dict = copy.deepcopy(settings_dict_in)
        settings_dict['seed'] = final_seed
        if settings_dict['seed'] >= 0:
            assert settings_dict_in['seed'] >= 0, (
                'if input seed was <0 it should not have been edited')
        if dyn_info['peak_start_ind'] != 0:
            settings_dict['nlive'] = ninit
        else:
            settings_dict['nlive'] = dyn_info['nlives_dict'][
                min(dyn_info['nlives_dict'].keys())]
        settings_dict['nlives'] = dyn_info['nlives_dict']
        # To write .ini files correctly, read_resume must be type bool not
        # np.bool
        settings_dict['read_resume'] = bool(dyn_info['peak_start_ind'] != 0)
        settings_dict['file_root'] = settings_dict_in['file_root'] + '_dyn'
    run_polychord(settings_dict, comm=comm)
    if rank == 0:
        # Step 4: process output and tidy
        # -------------------------------
        # Combine initial and dynamic runs
        run = dyPolyChord.output_processing.process_dypolychord_run(
            settings_dict_in['file_root'], settings_dict_in['base_dir'],
            dynamic_goal=dynamic_goal, logl_warn_only=logl_warn_only)
        # Save combined output in PolyChord format
        nestcheck.write_polychord_output.write_run_output(
            run, logl_warn_only=logl_warn_only,
            stats_means_errs=stats_means_errs, **output_settings)
        if clean:
            # Remove temporary files
            clean_extra_output(root_name)


# Helper functions
# ----------------

def clean_extra_output(root_name):
    """Clean the additional output files made by dyPolyChord, leaving only
    output files for the combined run in PolyChord format.

    Parameters
    ----------
    root_name: str
        File root. Equivalent to os.path.join(base_dir, file_root).
    """
    os.remove(root_name + '_dyn_info.pkl')
    for extra in ['init', 'dyn']:
        os.remove(root_name + '_{0}.stats'.format(extra))
        os.remove(root_name + '_{0}_dead-birth.txt'.format(extra))
        os.remove(root_name + '_{0}_dead.txt'.format(extra))
        # tidy up remaining .resume files (if the function has reach this
        # point, both the initial and dynamic runs have finished so we
        # shouldn't need to resume)
        try:
            os.remove(root_name + '_{0}.resume'.format(extra))
        except OSError:
            pass


def check_settings(settings_dict_in):
    """
    Checks the input dictionary of PolyChord settings. Issues warnings where
    these are not appropriate, and adds default values.

    Parameters
    ----------
    settings_dict_in: dict
        PolyChord settings to use.

    Returns
    -------
    settings_dict: dict
        Updated settings dictionary including default and mandatory values.
    output_settings: dict
        Settings for writing output files which are saved until the final
        output files are calculated at the end.
    """
    default_settings = {'nlive': 100,
                        'num_repeats': 20,
                        'file_root': 'temp',
                        'base_dir': 'chains',
                        'seed': -1,
                        'do_clustering': True,
                        'max_ndead': -1,
                        'equals': True,
                        'posteriors': True}
    mandatory_settings = {'nlives': {},
                          'write_dead': True,
                          'write_stats': True,
                          'write_paramnames': False,
                          'write_prior': False,
                          'write_live': False,
                          'write_resume': False,
                          'read_resume': False,
                          'cluster_posteriors': False,
                          'boost_posterior': 0.0}
    settings_dict = copy.deepcopy(settings_dict_in)
    # assign default settings
    for key, value in default_settings.items():
        if key not in settings_dict:
            settings_dict[key] = value
    # Produce warning if settings_dict_in has different values for any
    # mandatory settings.
    for key, value in mandatory_settings.items():
        if key in settings_dict_in and settings_dict_in[key] != value:
            warnings.warn((
                'dyPolyChord currently only allows the setting {0}={1}, '
                'so I am proceeding with this. You tried to specify {0}={2}.'
                .format(key, value, settings_dict_in[key])), UserWarning)
        settings_dict[key] = value
    # Extract output settings (not needed until later)
    output_settings = {}
    for key in ['posteriors', 'equals']:
        output_settings[key] = settings_dict[key]
        settings_dict[key] = False
    return settings_dict, output_settings


def run_and_save_resumes(run_polychord, settings_dict_in, init_step,
                         seed_increment, comm=None):
    """
    Run PolyChord pausing after every init_step dead points to save a resume
    file.

    Parameters
    ----------
    run_polychord: callable
        Callable which runs PolyChord with the desired likelihood and prior,
        and takes a settings dictionary as its argument.
    settings_dict: dict
        PolyChord settings to use (see check_settings for information on
        allowed and default settings).
    ninit_step: int, optional
        Number of samples taken between saving .resume files in Step 1.
    seed_increment: int, optional
        If seeding is used (PolyChord seed setting >= 0), this increment is
        added to PolyChord's random seed each time it is run to avoid
        repeated points.
        When running in parallel using MPI, PolyChord hashes the seed with the
        MPI rank using IEOR. Hence you need seed_increment to be > number of
        processors to ensure no two processes use the same seed.
        When running repeated results you need to increment the seed used for
        each run by some number >> seed_increment.
    comm: None or mpi4py MPI.COMM object, optional
        For MPI parallelisation.

    Returns
    -------
    step_ndead: list of ints
        Numbers of dead points at which resume files are saved.
    resume_outputs: dict
        Dictionary containing run output (contents of .stats file) at each
        resume. Keys are elements of step_ndead.
    seed: int
        Random seed. This is incremented after each run so it can be used
        when resuming without generating correlated points.
    """
    # set up rank if running with MPI
    settings_dict = copy.deepcopy(settings_dict_in)
    if comm is not None:
        rank = comm.Get_rank()
    else:
        rank = 0
    if rank == 0:
        root_name = os.path.join(settings_dict['base_dir'],
                                 settings_dict['file_root'])
        try:
            os.remove(root_name + '.resume')
        except OSError:
            pass
        settings_dict['write_resume'] = True
        settings_dict['read_resume'] = True
        step_ndead = []
        resume_outputs = {}
    add_points = True
    while add_points:
        if rank == 0:
            settings_dict['max_ndead'] = (len(step_ndead) + 1) * init_step
        run_polychord(settings_dict, comm=comm)
        if rank == 0:
            if settings_dict['seed'] >= 0:
                settings_dict['seed'] += seed_increment
            run_output = nestcheck.data_processing.process_polychord_stats(
                settings_dict['file_root'], settings_dict['base_dir'])
            # Store run outputs for getting number of likelihood calls
            # while accounding for resuming a run.
            resume_outputs[run_output['ndead']] = run_output
            step_ndead.append(run_output['ndead'] - settings_dict['nlive'])
            if len(step_ndead) >= 2 and step_ndead[-1] == step_ndead[-2]:
                add_points = False
            # store resume file in new file path
            shutil.copyfile(
                root_name + '.resume',
                root_name + '_' + str(step_ndead[-1]) + '.resume')
        if comm is not None:
            add_points = comm.bcast(add_points, root=0)
    return step_ndead, resume_outputs, settings_dict['seed']



#!/usr/bin/env python
"""
Loglikelihood functions for use with PyPolyChord (PolyChord's python wrapper).

PolyChord v1.14 requires likelihoods to be callables with parameter and return
types:

Parameters
----------
theta: float or 1d numpy array

Returns
-------
logl: float
    Loglikelihood.
phi: list of length nderived
    Any derived parameters.

We use classes with the loglikelihood defined in the __call__ property, as
this provides convenient way of storing other information such as
hyperparameter values. These objects can be used in the same way as functions
due to python's "duck typing" (alternatively you can define likelihoods
using functions).
"""
import copy
import numpy as np
import scipy.special


class Gaussian(object):

    """Symmetric Gaussian loglikelihood centered on the origin."""

    def __init__(self, sigma=1.0, nderived=0):
        """
        Set up likelihood object's hyperparameter values.

        Parameters
        ----------
        sigma: float, optional
            Standard deviation of Gaussian (the same for each parameter).
        nderived: int, optional
            Number of derived parameters.
        """
        self.sigma = sigma
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        logl = -(np.sum(theta ** 2) / (2 * self.sigma ** 2))
        logl -= np.log(2 * np.pi * (self.sigma ** 2)) * len(theta) / 2.0
        return logl, [0.0] * self.nderived


class GaussianShell(object):

    """Gaussian Shell loglikelihood centred on the origin."""

    def __init__(self, sigma=0.2, rshell=2, nderived=0):
        """
        Set up likelihood object's hyperparameter values.

        Parameters
        ----------
        sigma: float, optional
            Standard deviation of Gaussian (the same for each parameter).
        rshell: float, optional
            Distance of shell peak from origin.
        nderived: int, optional
            Number of derived parameters.
        """
        self.sigma = sigma
        self.rshell = rshell
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        rad = np.sqrt(np.sum(theta ** 2))
        logl = - ((rad - self.rshell) ** 2) / (2 * self.sigma ** 2)
        return logl, [0.0] * self.nderived


class Rastrigin(object):

    """Rastrigin loglikelihood as described in the PolyChord paper."""

    def __init__(self, a=10, nderived=0):
        """
        Set up likelihood object's hyperparameter values.

        Parameters
        ----------
        a: float, optional
        nderived: int, optional
            Number of derived parameters.
        """
        self.a = a
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        ftheta = self.a * len(theta)
        for th in theta:
            ftheta += (th ** 2) - self.a * np.cos(2 * np.pi * th)
        logl = -ftheta
        return logl, [0.0] * self.nderived


class Rosenbrock(object):

    """Rosenbrock loglikelihood as described in the PolyChord paper."""

    def __init__(self, a=1, b=100, nderived=0):
        """
        Define likelihood hyperparameter values.

        Parameters
        ----------
        theta: 1d numpy array
            Parameters.
        a: float, optional
        b: float, optional
        nderived: int, optional
            Number of derived parameters.
        """
        self.a = a
        self.b = b
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        ftheta = 0
        for i in range(len(theta) - 1):
            ftheta += (self.a - theta[i]) ** 2
            ftheta += self.b * ((theta[i + 1] - (theta[i] ** 2)) ** 2)
        logl = -ftheta
        return logl, [0.0] * self.nderived


class GaussianMix(object):

    """
    Gaussian mixture likelihood in :math:`\\ge 2` dimensions with up to
    4 compoents.

    Each component has the same standard deviation :math:`\\sigma`, and
    they their centres respectively have :math:`(\\theta_1, \\theta_2)`
    coordinates:

    (0, sep), (0, -sep), (sep, 0), (-sep, 0).
    """

    def __init__(self, sep=4, weights=(0.4, 0.3, 0.2, 0.1), sigma=1,
                 nderived=0):
        """
        Define likelihood hyperparameter values.

        Parameters
        ----------
        sep: float
            Distance from each Gaussian to the origin.
        weights: iterable of floats
            weights of each Gaussian component.
        sigma: float
            Stanard deviation of Gaussian components.
        nderived: int, optional
            Number of derived parameters.
        """
        assert len(weights) in [2, 3, 4], (
            'Weights must have 2, 3 or 4 components. Weights=' + str(weights))
        assert np.isclose(sum(weights), 1), (
            'Weights must sum to 1! Weights=' + str(weights))
        self.nderived = nderived
        self.weights = weights
        self.sigmas = [sigma] * len(weights)
        positions = []
        positions.append(np.asarray([0, sep]))
        positions.append(np.asarray([0, -sep]))
        positions.append(np.asarray([sep, 0]))
        positions.append(np.asarray([-sep, 0]))
        self.positions = positions[:len(weights)]

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        thetas = []
        for pos in self.positions:
            thetas.append(copy.deepcopy(theta))
            thetas[-1][:2] -= pos
        logls = [(Gaussian(sigma=self.sigmas[i])(thetas[i])[0]
                  + np.log(self.weights[i])) for i in range(len(self.weights))]
        logl = scipy.special.logsumexp(logls)
        return logl, [0.0] * self.nderived



#!/usr/bin/env python
"""
Python priors for use with PolyChord.

PolyChord v1.14 requires priors to be callables with parameter and return
types:

Parameters
----------
hypercube: float or 1d numpy array
    Parameter positions in the prior hypercube.

Returns
-------
theta: float or 1d numpy array
    Corresponding physical parameter coordinates.

Input hypercube values numpy array is mapped to physical space using the
inverse CDF (cumulative distribution function) of each parameter.
See the PolyChord papers for more details.

We use classes with the prior defined in the __call__ property, as
this provides convenient way of storing other information such as
hyperparameter values. The objects be used in the same way as functions
due to python's "duck typing" (or alternatively you can just define prior
functions).
"""
import numpy as np
import scipy


class Uniform(object):

    """Uniform prior."""

    def __init__(self, minimum=0.0, maximum=1.0):
        """
        Set up prior object's hyperparameter values.

        Prior is uniform in [minimum, maximum] in each parameter.

        Parameters
        ----------
        hypercube: list of floats
            Point coordinate on unit hypercube (in probabily space).
            See the PolyChord papers for more details.
        minimum: float
        maximum: float
        """
        assert maximum > minimum
        self.maximum = maximum
        self.minimum = minimum

    def __call__(self, hypercube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        hypercube: list of floats
            Point coordinate on unit hypercube (in probabily space).
            See the PolyChord papers for more details.

        Returns
        -------
        theta: list of floats
            Physical parameter values corresponding to hypercube.
        """
        return self.minimum + (self.maximum - self.minimum) * hypercube


class Gaussian(object):

    """Symmetric Gaussian prior centred on the origin."""

    def __init__(self, sigma=10.0):
        """
        Set up prior object's hyperparameter values.

        Parameters
        ----------
        sigma: float
            Standard deviation of Gaussian prior in each parameter.
        """
        self.sigma = sigma

    def __call__(self, hypercube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        hypercube: list of floats
            Point coordinate on unit hypercube (in probabily space).
            See the PolyChord papers for more details.

        Returns
        -------
        theta: list of floats
            Physical parameter values corresponding to hypercube.
        """
        theta = scipy.special.erfinv(2 * hypercube - 1)
        theta *= self.sigma * np.sqrt(2)
        return theta
